//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/DiscreteAxis.cpp
//! @brief     Implements class DiscreteAxis.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/DiscreteAxis.h"
#include "Base/Axis/Bin.h"
#include "Base/Util/VectorUtil.h"
#include <algorithm>
#include <iomanip>
#include <stdexcept>

const size_t min_axis_size = 2;

DiscreteAxis::DiscreteAxis(const std::string& name, const std::vector<double>& coordinate_values)
    : IAxis(name)
    , m_coordinates(coordinate_values)
{
    sanityCheck();
}

DiscreteAxis::DiscreteAxis(const std::string& name, size_t n, double first, double last)
    : DiscreteAxis(name, VectorUtil::make_grid(n, first, last))
{
}

DiscreteAxis* DiscreteAxis::clone() const
{
    return new DiscreteAxis(axisName(), m_coordinates);
}

Bin1D DiscreteAxis::bin(size_t index) const
{
    checkIndex(index);
    return {minary(index), maxary(index)};
}

double DiscreteAxis::min() const
{
    return minary(0);
}

double DiscreteAxis::max() const
{
    return maxary(m_coordinates.size() - 1);
}

double DiscreteAxis::binCenter(size_t index) const
{
    checkIndex(index);
    return m_coordinates[index];
}

size_t DiscreteAxis::findClosestIndex(double value) const
{
    if (value <= m_coordinates.front())
        return 0;
    if (value >= m_coordinates.back())
        return m_coordinates.size() - 1;

    const auto begin = m_coordinates.begin();
    auto result = std::lower_bound(begin, m_coordinates.end(), value);
    auto index = static_cast<size_t>(std::distance(begin, result));
    return value < minary(index) ? index - 1 : index;
}

std::vector<double> DiscreteAxis::binBoundaries() const
{
    std::vector<double> result;
    const size_t v_size = m_coordinates.size();
    result.reserve(v_size + 1);
    for (size_t i = 0; i < v_size; ++i)
        result.push_back(minary(i));
    result.push_back(maxary(v_size - 1));
    return result;
}

void DiscreteAxis::clip(double lower, double upper)
{
    if (lower >= upper)
        throw std::runtime_error("DiscreteAxis::clip() called with invalid lower >= upper");

    using diff_t = std::vector<double>::iterator::difference_type;
    const auto begin = m_coordinates.begin() + static_cast<diff_t>(findClosestIndex(lower));
    const auto end = m_coordinates.begin() + static_cast<diff_t>(findClosestIndex(upper)) + 1;

    m_coordinates = std::vector<double>(begin, end);
    sanityCheck();
}

void DiscreteAxis::print(std::ostream& ostr) const
{
    auto precision = std::setprecision(std::numeric_limits<double>::digits10 + 2);
    ostr << "DiscreteAxis(\"" << axisName() << "\", "
         << ", [";
    for (size_t i = 0, fin = m_coordinates.size() - 1; i < fin; ++i)
        ostr << precision << m_coordinates[i] << ",";
    ostr << precision << m_coordinates.back() << "])";
}

bool DiscreteAxis::equals(const IAxis& other) const
{
    if (!IAxis::equals(other))
        return false;
    if (const auto* otherAxis = dynamic_cast<const DiscreteAxis*>(&other))
        return m_coordinates == otherAxis->binCenters();
    return false;
}

double DiscreteAxis::minary(size_t index) const
{
    if (index == 0)
        return m_coordinates.front();
    return 0.5 * (m_coordinates[index] + m_coordinates[index - 1]);
}

double DiscreteAxis::maxary(size_t index) const
{
    if (index + 1 == m_coordinates.size())
        return m_coordinates.back();
    return 0.5 * (m_coordinates[index] + m_coordinates[index + 1]);
}

void DiscreteAxis::checkIndex(size_t index) const
{
    if (m_coordinates.size() > index)
        return;
    std::string message = "Error in DiscreteAxis::binCenter: passed index ";
    message += std::to_string(index) + " exceeds the size ";
    message += std::to_string(m_coordinates.size()) + " of the axis";
    throw std::runtime_error(message);
}

void DiscreteAxis::sanityCheck() const
{
    if (m_coordinates.size() < min_axis_size)
        throw std::runtime_error(
            "Error in DiscreteAxis::DiscreteAxis: the size of passed coordinate array is "
            "less than minimum acceptable value");

    const auto begin = m_coordinates.begin();
    const auto end = m_coordinates.end();

    if (!std::is_sorted(begin, end))
        throw std::runtime_error("Error in DiscreteAxis::DiscreteAxis: passed coordinates are "
                                 "not sorted in ascending order");

    if (std::adjacent_find(begin, end) != end)
        throw std::runtime_error("Error in DiscreteAxis::DiscreteAxis: passed coordinate vector "
                                 "contains repeating values");
}
