//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Specular/ComputeFluxMagnetic.cpp
//! @brief     Implements class SpecularMagneticStrategy.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2020
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Specular/ComputeFluxMagnetic.h"
#include "Base/Const/PhysicalConstants.h"
#include "Base/Util/Assert.h"
#include "Resample/Flux/MatrixFlux.h"
#include "Resample/Slice/KzComputation.h"
#include "Resample/Slice/SliceStack.h"
#include "Resample/Specular/TransitionMagneticNevot.h"
#include "Resample/Specular/TransitionMagneticTanh.h"
#include "Sample/Interface/LayerRoughness.h"
#include <algorithm>

namespace {

R3 differenceField(const SliceStack& slices, size_t i, bool forward)
{
    return (forward ? +1 : -1) * (slices[i].bField() - slices[0].bField());
}

std::pair<SpinMatrix, SpinMatrix> computeBackwardsSubmatrices(const MatrixFlux& coeff_i,
                                                              const MatrixFlux& coeff_i1,
                                                              double sigma,
                                                              const RoughnessModel& r_model)
{
    if (r_model == RoughnessModel::NEVOT_CROCE)
        return Compute::MagneticNevotCroceTransition::backwardsSubmatrices(coeff_i, coeff_i1,
                                                                           sigma);
    return Compute::TransitionMagneticTanh::backwardsSubmatrices(coeff_i, coeff_i1, sigma);
}

std::vector<MatrixFlux> computeFlux(const SliceStack& slices, const std::vector<complex_t>& kzs,
                                    const RoughnessModel& r_model, bool forward)
{
    const size_t N = slices.size();
    ASSERT(N > 0);
    ASSERT(kzs.size() == N);

    std::vector<MatrixFlux> result;
    result.reserve(N);

    const int kz0sign = kzs[0].real() >= 0 ? 1 : -1; // save sign to restore it later

    result.emplace_back(MatrixFlux(kz0sign, kzs[0], R3()));
    for (size_t i = 1; i < slices.size(); ++i)
        result.emplace_back(MatrixFlux(kz0sign, kzs[i], differenceField(slices, i, forward)));

    if (N == 1) {
        result[0].setT(SpinMatrix::One());
        result[0].setR(SpinMatrix());
        return result;
    }

    if (kzs[0] == 0.) {
        result[0].setT(SpinMatrix::One());
        result[0].setR(SpinMatrix::Diag(-1, -1));
        for (size_t i = 1; i < N; ++i) {
            result[i].setT(SpinMatrix());
            result[i].setR(SpinMatrix());
        }
        return result;
    }
    // This was the last special case. Now the regular case.

    std::vector<SpinMatrix> SMatrices(N - 1);
    std::vector<complex_t> Normalization(N - 1);

    // bottom boundary condition
    result.back().setT(SpinMatrix::One());
    result.back().setR(SpinMatrix());

    for (size_t i = N - 1; i > 0; --i) {
        double sigma = 0.;
        if (const auto* const roughness = slices.bottomRoughness(i))
            sigma = roughness->sigma();

        // compute the 2x2 submatrices in the write-up denoted as P+, P- and delta
        const auto [mp, mm] = computeBackwardsSubmatrices(result[i - 1], result[i], sigma, r_model);

        const SpinMatrix delta = result[i - 1].computeDeltaMatrix(slices[i - 1].thicknessOr0());

        // compute the rotation matrix
        SpinMatrix Si = mp + mm * result[i].R();
        SpinMatrix S(Si.d, -Si.b, -Si.c, Si.a);
        const complex_t norm = S.determinant();
        S = S * delta;

        // store the rotation matrix and normalization constant in order to rotate
        // the coefficients for all lower slices at the end of the computation
        SMatrices[i - 1] = S;
        Normalization[i - 1] = norm;

        // compute the reflection matrix and
        // rotate the polarization such that we have pure incoming states (T = I)
        S /= norm;

        // T is always equal to the identity at this point, no need to store
        result[i - 1].setR(delta * (mm + mp * result[i].R()) * S);
    }

    // now correct all amplitudes in forward direction by dividing with the remaining
    // normalization constants. In addition rotate the polarization by the amount
    // that was rotated above the current interface
    // if the normalization overflows, all amplitudes below that point are set to zero
    complex_t dumpingFactor = 1;
    SpinMatrix S = SpinMatrix::One();
    for (size_t i = 1; i < N; ++i) {
        dumpingFactor = dumpingFactor * Normalization[i - 1];
        S = SMatrices[i - 1] * S;

        if (std::isinf(std::norm(dumpingFactor))) {
            std::for_each(result.begin() + i, result.end(), [](auto& result) {
                result.setT(SpinMatrix());
                result.setR(SpinMatrix());
            });
            break;
        }

        result[i].setT(S / dumpingFactor); // T * S omitted, since T is always I
        result[i].setR(result[i].R() * S / dumpingFactor);
    }

    return result;
}

} // namespace

//  ************************************************************************************************
//  namespace implementation
//  ************************************************************************************************

Fluxes Compute::SpecularMagnetic::fluxes(const SliceStack& slices, const R3& k, bool forward)
{
    const bool top_exit = k.z() <= 0; // true if source or detector pixel are at z>=0
    if (slices.size() > 1 && !top_exit)
        throw std::runtime_error(
            "source or detector below horizon not yet implemented for polarized scattering"
            " (https://jugit.fz-juelich.de/mlz/bornagain/-/issues/101)");

    std::vector<complex_t> kz = Compute::Kz::computeReducedKz(slices, k);
    ASSERT(kz.size() == slices.size());

    std::vector<MatrixFlux> fluxes = computeFlux(slices, kz, slices.roughnessModel(), forward);
    // copy to other vector type:
    Fluxes result;
    for (auto& flux : fluxes)
        result.emplace_back(new MatrixFlux(flux));
    return result;
}

SpinMatrix Compute::SpecularMagnetic::topLayerR(const SliceStack& slices,
                                                const std::vector<complex_t>& kzs, bool forward)
{
    const auto N = slices.size();
    ASSERT(kzs.size() == N);

    if (N == 1)
        return {};
    if (kzs[0] == 0.)
        return -SpinMatrix::One();

    const int kz0sign = kzs[0].real() >= 0 ? 1 : -1; // save sign to restore it later

    MatrixFlux c_i1(kz0sign, kzs[N - 1], differenceField(slices, N - 1, forward));

    // bottom boundary condition
    c_i1.setR(SpinMatrix());

    for (size_t i = N - 1; i > 0; --i) {
        MatrixFlux c_i(kz0sign, kzs[i - 1], differenceField(slices, i - 1, forward));

        double sigma = 0.;
        if (const auto* const roughness = slices.bottomRoughness(i - 1))
            sigma = roughness->sigma();

        // compute the 2x2 submatrices in the write-up denoted as P+, P- and delta
        const auto [mp, mm] =
            computeBackwardsSubmatrices(c_i, c_i1, sigma, slices.roughnessModel());

        const SpinMatrix delta = c_i.computeDeltaMatrix(slices[i - 1].thicknessOr0());

        // compute the rotation matrix
        SpinMatrix Si = mp + mm * c_i1.R();
        SpinMatrix S(Si.d, -Si.b, -Si.c, Si.a);
        const complex_t norm = S.determinant();
        S = S * (delta / norm);

        c_i.setR(delta * (mm + mp * c_i1.R()) * S);
        c_i1 = c_i;
    }
    return c_i1.R();
}
